/*************************************************** 
  Si4713-Based Guitar FM Transmitter
  Jonathan Rico
  
  Uses code from:
  AdaFruit_Si4713 : https://github.com/adafruit/Adafruit-Si4713-Library

 ****************************************************/

#include <EEPROM.h>
#include <Wire.h>
#include <Adafruit_Si4713.h>
#include "ledDisplay.h"

#define OVERDRIVE_PIN 4 //Preamp bypass signal
#define RESETPIN 9 // Si4713 !RST pin
#define DISPLAYENABLE 8 // Pin used to blank display
#define DISPSLEEPTIME 10000 //Display sleep timeout
// Interface tactile switches
#define SW_A 2
#define SW_B 3
#define SW_C 10

#define FMTXPOWER 115 // dBuV, 88-115 max
#define INPUTLEVEL 0x327C // Input level register value
#define BUTTON_HOLD_TIME 500 // Time in ms after which the button is considered being held
#define REPEATSPEED 150 // Time between each button repeat
#define TEXTSPEED 150 // Time between each character shift for long text string
#define POT_ADDRESS 0x2F // Address of digital potentiometer


#define SERIALDEBUG // Enable debug output
//#define BUTTONDEBUG // Prints button state
#define RADIO_ON // Enable transmitter functionnality

Adafruit_Si4713 radio = Adafruit_Si4713(RESETPIN);

uint16_t tuningFreq = 1023; //Si4713 FM Transmitting frequency
byte ampLevel = 77; // Preamp gain level
char RDSString[] = "GuitarFM"; //Si4713 RDS String (radio name)


void setup() {

  //Display pins setup
  pinMode(ch1, OUTPUT);
  pinMode(ch2, OUTPUT);
  pinMode(ch3, OUTPUT);
  pinMode(ch4, OUTPUT);
  pinMode(latchPin, OUTPUT);
  pinMode(DISPLAYENABLE, OUTPUT);
  clearLedDisplay();

  // Interface buttons set-up
  pinMode(SW_A, INPUT);
  pinMode(SW_B, INPUT);
  pinMode(SW_C, INPUT);

  //Preamp setup
  pinMode(OVERDRIVE_PIN, OUTPUT);
  digitalWrite(OVERDRIVE_PIN, LOW); // Disable diode overdrive (HIGH to enable)

  
  // Retrieve saved settings from memory
  if(EEPROM.read(4) == 42) {
    byte overdrive = 0;

    tuningFreq = (EEPROM.read(0) * 256) + EEPROM.read(1); 
    ampLevel = EEPROM.read(2);
    overdrive = EEPROM.read(3);
    digitalWrite(OVERDRIVE_PIN, overdrive);
  }

  //Serial init
  #ifdef SERIALDEBUG
  Serial.begin(9600);
  Serial.println("GuitarFM - Testing");
  #endif

  digitalWrite(DISPLAYENABLE, LOW);  // Turn on display
  dispString("rock n roll", TEXTSPEED);
  blinkLedDisplay(1000);

  // Si4713 I2C Init
  #ifdef RADIO_ON
  while (!radio.begin(0x11)) { // begin with address 0x11 (CS low)
    dispNumber(0000,0);
    #ifdef SERIALDEBUG
    Serial.println("FM chip not responding, retrying...");
    dispString("tx error", TEXTSPEED);
    #endif
  }
  #endif


  //Si4713 Power settings
  #ifdef SERIALDEBUG
  Serial.print("\nSet TX power");
  #endif
  #ifdef RADIO_ON
  radio.setTXpower(FMTXPOWER);  // dBuV, 88-115 max
  #endif

  //Si4713 Input settings
  #ifdef SERIALDEBUG
  Serial.print("\nSet Input level");
  #endif
  // Sets input resistance to 60k
  #ifdef RADIO_ON
  radio.setProperty(SI4713_PROP_TX_LINE_LEVEL_INPUT_LEVEL, INPUTLEVEL);
  #endif




  //Si4713 FM frequency setup
  #ifdef SERIALDEBUG
  Serial.print("\nTuning into "); 
  Serial.print(tuningFreq/10); 
  Serial.print('.'); 
  Serial.println(tuningFreq % 10);
  #endif
  #ifdef RADIO_ON
  radio.tuneFM((tuningFreq * 10)); //Actual tuning
  #endif
  
  //Si4713 status read
  #ifdef RADIO_ON
  radio.readTuneStatus();//Retrieve current chip status
  #endif
  #ifdef SERIALDEBUG
  Serial.print("\tCurr freq: "); 
  Serial.println(radio.currFreq);
  Serial.print("\tCurr dBuV:"); 
  Serial.println(radio.currdBuV);
  Serial.print("\tCurr ANTcap:"); 
  Serial.println(radio.currAntCap);
  #endif
  
  //Si4713 RDS transmission (radio name) setup
  #ifdef RADIO_ON
  radio.beginRDS();
  radio.setRDSstation(RDSString);
  #endif
  #ifdef SERIALDEBUG
  Serial.print(RDSString);
  Serial.println(" now live !");  
  #endif

  dispNumber(1023,3);


}

void sleepDisplay() {
  loadingAnimation();
}

void setAmpGain(int gain) {

  static byte pot_value = 100;

  pot_value = map(gain, 0, 100, 0, 127);

  Wire.beginTransmission(POT_ADDRESS);
  Wire.write(pot_value);
  Wire.endTransmission();

  #ifdef SERIALDEBUG
  Serial.print("Amp gain set :");
  Serial.println(gain, DEC);
  #endif
}

void saveSettings(void) {
  byte overdrive = digitalRead(OVERDRIVE_PIN);

  EEPROM.write(0, (tuningFreq/256));
  EEPROM.write(1, (tuningFreq%256));
  EEPROM.write(2, ampLevel);
  EEPROM.write(3, overdrive);
  EEPROM.write(4, 42);

  #ifdef SERIALDEBUG
  Serial.println("Settings saved to EEPROM");
  #endif
}

void refreshInterface(void) {
  #define AMP 0
  #define FREQ 1
  #define DIST 2
  #define DIST_ON 1
  #define DIST_OFF 0
  #define SETTING_TIME 500

  char settingLabels[3][10] = {
    {'a', 'm', 'p', '\0'},
    {'f', 'r', 'e', 'q', '\0'},
    {'d', 'i', 's', 't', '\0'}
  };

  /*
  char settingLabels[3][10] = {
    {'g', 'a', 'i', 'n', '\0'},
    {'t', 'u', 'n', 'i', 'n', 'g', '\0'},
    {'o', 'v', 'e', 'r', 'd', 'r', 'i', 'v', 'e', '\0'}
  };
  */

  static byte currentSetting = FREQ;
  static byte prevSetting = FREQ;
  static bool firstTime = 1;

  unsigned long startTime = millis();
  static unsigned long lastSettingChange = millis();
  
  //refreshDisplayDelay(150);
  while(!(digitalRead(SW_B))) {
    refreshDisplayDelay(200);
    if((millis() - startTime) > BUTTON_HOLD_TIME) {
      saveSettings();
      blinkLedDisplay(1000);
      break;
    } 
    else if(digitalRead(SW_B)) {
      currentSetting++;
      if(currentSetting > DIST) currentSetting = AMP;
    }
  }

  if(currentSetting != prevSetting) {
    #ifdef SERIALDEBUG
      Serial.print("Changed setting : ");
      Serial.println(settingLabels[currentSetting]);
    #endif

    firstTime = 1;
    dispString(settingLabels[currentSetting]);
    refreshDisplayDelay(SETTING_TIME);
    lastSettingChange = millis();
    prevSetting = currentSetting;
  }

  if(((millis() - lastSettingChange) < DISPSLEEPTIME) || !digitalRead(SW_A) || !digitalRead(SW_B) || !digitalRead(SW_C)) {

    switch (currentSetting) {
        case AMP :
          dispNumber(ampLevel, 0);
          startTime = millis();
          while(!(digitalRead(SW_A))) {
            if((millis() - startTime) > BUTTON_HOLD_TIME) {
              if(ampLevel > 1) ampLevel--;
            }
            else {
              if(ampLevel >= 1) ampLevel--;
              refreshDisplayDelay(REPEATSPEED);
            }
            dispNumber(ampLevel, 0);
            refreshDisplay();
            lastSettingChange = millis();
            setAmpGain(ampLevel);
          }
          
          while(!(digitalRead(SW_C))) {
            if((millis() - startTime) > BUTTON_HOLD_TIME) {
              if(ampLevel < 100) ampLevel++;
            }
            else {
              if(ampLevel < 100) ampLevel++;
              refreshDisplayDelay(REPEATSPEED);
            }
            dispNumber(ampLevel, 0);
            refreshDisplay();
            lastSettingChange = millis();
            setAmpGain(ampLevel);
          }
          
          break;
        case FREQ :

          dispNumber(tuningFreq, 3);
          static int oldFreq = 0;

          startTime = millis();
          while(!(digitalRead(SW_A))) {

            if((millis() - startTime) > BUTTON_HOLD_TIME) {
              if(tuningFreq > 880) tuningFreq--;
            }
            else {
              if(tuningFreq > 880) tuningFreq--;
              refreshDisplayDelay(REPEATSPEED);
            }

            dispNumber(tuningFreq, 3);
            lastSettingChange = millis();
            refreshDisplay();
          }
          
          while(!(digitalRead(SW_C))) {

            if((millis() - startTime) > BUTTON_HOLD_TIME) {
              if(tuningFreq < 1080) tuningFreq++;
            }
            else {
              if(tuningFreq < 1080) tuningFreq++;
              refreshDisplayDelay(REPEATSPEED);
            }

            dispNumber(tuningFreq, 3);
            lastSettingChange = millis();
            refreshDisplay();
          }

          if(tuningFreq != oldFreq) {

            #ifdef SERIALDEBUG
            Serial.print("\nTuning into "); 
            Serial.print(tuningFreq/10); 
            Serial.print('.'); 
            Serial.println(tuningFreq % 10);
            #endif

            #ifdef RADIO_ON
            radio.tuneFM((tuningFreq * 10)); //Actual tuning
            #endif

            oldFreq = tuningFreq;
          }
          break;
        case DIST :
          static byte overdrive = digitalRead(OVERDRIVE_PIN);

          if(!(digitalRead(SW_A))) {
            overdrive = DIST_OFF;
            lastSettingChange = millis();
          }
          else if(!(digitalRead(SW_C))) {
            overdrive = DIST_ON;
            lastSettingChange = millis();
          }
          if((digitalRead(OVERDRIVE_PIN) != overdrive) || firstTime) {
            firstTime = 0;
            if(overdrive==DIST_ON) 
              dispString("on");
            else if (overdrive==DIST_OFF) 
              dispString("off");
            digitalWrite(OVERDRIVE_PIN, overdrive);

            #ifdef SERIALDEBUG
             Serial.print("Overdrive : ");
             Serial.println(overdrive, DEC);
            #endif
          }
          break;
        default:
          break;
    }
  }
  else sleepDisplay();

  #ifdef BUTTONDEBUG
    if(!(digitalRead(SW_A))) {
      Serial.println("Button A pressed");  
    }
    if(!(digitalRead(SW_B))) {
      Serial.println("Button B pressed");  
    }
    if(!(digitalRead(SW_C))) {
      Serial.println("Button C pressed");  
    }
  #endif
}

void loop() {
  refreshDisplay(); //Refresh POV 4-digit 8-segment LED display
  refreshInterface(); // Refresh interface (read buttons etc..)

}

